import warnings


def __getattr__(name):
    if name == "NotDirectoryError":
        warnings.warn(
            "upath.errors.NotDirectoryError is deprecated. "
            "Use NotADirectoryError instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return NotADirectoryError
    raise AttributeError(name)



import sys
import pytest


def skip_on_windows(func):
    return pytest.mark.skipif(
        sys.platform.startswith("win"), reason="Don't run on Windows"
    )(func)


def only_on_windows(func):
    return pytest.mark.skipif(
        not sys.platform.startswith("win"), reason="Only run on Windows"
    )(func)


def posixify(path):
    return str(path).replace("\\", "/")



from __future__ import annotations

import pathlib
import re
import sys
from os import PathLike
from pathlib import _PosixFlavour  # type: ignore
from typing import Sequence
from typing import TypeVar
from typing import TYPE_CHECKING
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

from fsspec.registry import get_filesystem_class
from fsspec.utils import stringify_path

from upath.registry import get_upath_class

if TYPE_CHECKING:
    from typing import Any
    from typing import Generator
    from urllib.parse import SplitResult
    from fsspec.spec import AbstractFileSystem

__all__ = [
    "UPath",
]


class _FSSpecAccessor:
    __slots__ = ("_fs",)

    def __init__(self, parsed_url: SplitResult | None, **kwargs: Any) -> None:
        if parsed_url:
            cls = get_filesystem_class(parsed_url.scheme)
            url_kwargs = cls._get_kwargs_from_urls(urlunsplit(parsed_url))
        else:
            cls = get_filesystem_class(None)
            url_kwargs = {}
        url_kwargs.update(kwargs)
        self._fs = cls(**url_kwargs)

    def _format_path(self, path: UPath) -> str:
        return path.path

    def open(self, path, mode="r", *args, **kwargs):
        return self._fs.open(self._format_path(path), mode, *args, **kwargs)

    def stat(self, path, **kwargs):
        return self._fs.stat(self._format_path(path), **kwargs)

    def listdir(self, path, **kwargs):
        p_fmt = self._format_path(path)
        contents = self._fs.listdir(p_fmt, **kwargs)
        if len(contents) == 0 and not self._fs.isdir(p_fmt):
            raise NotADirectoryError
        elif (
            len(contents) == 1
            and contents[0]["name"] == p_fmt
            and contents[0]["type"] == "file"
        ):
            raise NotADirectoryError
        return contents

    def glob(self, _path, path_pattern, **kwargs):
        return self._fs.glob(self._format_path(path_pattern), **kwargs)

    def exists(self, path, **kwargs):
        return self._fs.exists(self._format_path(path), **kwargs)

    def info(self, path, **kwargs):
        return self._fs.info(self._format_path(path), **kwargs)

    def rm(self, path, recursive, **kwargs):
        return self._fs.rm(
            self._format_path(path), recursive=recursive, **kwargs
        )

    def mkdir(self, path, create_parents=True, **kwargs):
        return self._fs.mkdir(
            self._format_path(path), create_parents=create_parents, **kwargs
        )

    def makedirs(self, path, exist_ok=False, **kwargs):
        return self._fs.makedirs(
            self._format_path(path), exist_ok=exist_ok, **kwargs
        )

    def touch(self, path, **kwargs):
        return self._fs.touch(self._format_path(path), **kwargs)


class _UriFlavour(_PosixFlavour):
    def parse_parts(self, parts):
        parsed = []
        sep = self.sep
        drv = root = ""
        it = reversed(parts)
        for part in it:
            if part:
                drv, root, rel = self.splitroot(part)
                if not root or root and rel:
                    for x in reversed(rel.split(sep)):
                        parsed.append(sys.intern(x))

        if drv or root:
            parsed.append(drv + root)
        parsed.reverse()
        return drv, root, parsed

    def splitroot(self, part, sep="/"):
        # Treat the first slash in the path as the root if it exists
        if part and part[0] == sep:
            return "", sep, part[1:]
        return "", "", part


PT = TypeVar("PT", bound="UPath")


class UPath(pathlib.Path):
    __slots__ = (
        "_url",
        "_kwargs",
        "_accessor",  # overwritten because of default in Python 3.10
    )
    _flavour = _UriFlavour()
    _default_accessor = _FSSpecAccessor

    # typing
    _drv: str
    _root: str
    _str: str
    _url: SplitResult | None
    _parts: list[str]
    _closed: bool
    _accessor: _FSSpecAccessor

    def __new__(cls: type[PT], *args: str | PathLike, **kwargs: Any) -> PT:
        args_list = list(args)
        other = args_list.pop(0)

        if isinstance(other, pathlib.Path):
            # Create a (modified) copy, if first arg is a Path object
            _cls: type[Any] = type(other)
            drv, root, parts = _cls._parse_args(args_list)
            drv, root, parts = _cls._flavour.join_parsed_parts(
                other._drv, other._root, other._parts, drv, root, parts  # type: ignore # noqa: E501
            )

            _kwargs = getattr(other, "_kwargs", {})
            _url = getattr(other, "_url", None)
            other_kwargs = _kwargs.copy()
            if _url:
                other_kwargs["url"] = _url
            new_kwargs = _kwargs.copy()
            new_kwargs.update(kwargs)

            return _cls(  # type: ignore
                _cls._format_parsed_parts(drv, root, parts, **other_kwargs),
                **new_kwargs,
            )

        else:
            url = stringify_path(other)
            parsed_url = urlsplit(url)
            for key in ["scheme", "netloc"]:
                val = kwargs.get(key)
                if val:
                    parsed_url = parsed_url._replace(**{key: val})

            upath_cls = get_upath_class(protocol=parsed_url.scheme)
            if upath_cls is None:
                # treat as local filesystem, return PosixPath or WindowsPath
                return pathlib.Path(*args, **kwargs)  # type: ignore

            else:
                # return upath instance
                args_list.insert(0, parsed_url.path)
                return upath_cls._from_parts(
                    args_list, url=parsed_url, **kwargs
                )

    def __getattr__(self, item: str) -> Any:
        if item == "_accessor":
            # cache the _accessor attribute on first access
            kwargs = self._kwargs.copy()
            self._accessor = _accessor = self._default_accessor(
                self._url, **kwargs
            )
            return _accessor
        else:
            raise AttributeError(item)

    def _make_child(self: PT, args: list[str]) -> PT:
        drv, root, parts = self._parse_args(args)
        drv, root, parts = self._flavour.join_parsed_parts(
            self._drv, self._root, self._parts, drv, root, parts
        )
        return self._from_parsed_parts(
            drv, root, parts, url=self._url, **self._kwargs
        )

    def _make_child_relpath(self: PT, part: str) -> PT:
        # This is an optimization used for dir walking.  `part` must be
        # a single part relative to this path.
        parts = self._parts + [part]
        return self._from_parsed_parts(
            self._drv, self._root, parts, url=self._url, **self._kwargs
        )

    @classmethod
    def _format_parsed_parts(
        cls: type[PT],
        drv: str,
        root: str,
        parts: list[str],
        url: SplitResult | None = None,
        **kwargs: Any,
    ) -> str:
        if parts:
            join_parts = parts[1:] if parts[0] == "/" else parts
        else:
            join_parts = []
        if drv or root:
            path: str = drv + root + cls._flavour.join(join_parts)
        else:
            path = cls._flavour.join(join_parts)
        if not url:
            scheme: str = kwargs.get("scheme", "file")
            netloc: str = kwargs.get("netloc", "")
        else:
            scheme, netloc = url.scheme, url.netloc
        scheme = scheme + ":"
        netloc = "//" + netloc if netloc else ""
        formatted = scheme + netloc + path
        return formatted

    @property
    def path(self) -> str:
        if self._parts:
            join_parts = (
                self._parts[1:] if self._parts[0] == "/" else self._parts
            )
            path: str = self._flavour.join(join_parts)
            return self._root + path
        else:
            return "/"

    def open(self, *args, **kwargs):
        return self._accessor.open(self, *args, **kwargs)

    @property
    def parent(self: PT) -> PT:
        """The logical parent of the path."""
        drv = self._drv
        root = self._root
        parts = self._parts
        if len(parts) == 1 and (drv or root):
            return self
        return self._from_parsed_parts(
            drv, root, parts[:-1], url=self._url, **self._kwargs
        )

    def stat(self):
        return self._accessor.stat(self)

    def samefile(self, other_path) -> bool:
        raise NotImplementedError

    def iterdir(self: PT) -> Generator[PT, None, None]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for name in self._accessor.listdir(self):
            # fsspec returns dictionaries
            if isinstance(name, dict):
                name = name.get("name")
            if name in {".", ".."}:
                # Yielding a path object for these makes little sense
                continue
            # only want the path name with iterdir
            name = self._sub_path(name)
            yield self._make_child_relpath(name)

    def relative_to(self: PT, *other: str | PathLike) -> PT:
        for other_item in other:
            if not isinstance(other_item, self.__class__) and not isinstance(
                other_item, str
            ):
                raise ValueError(
                    f"{repr(self)} and {repr(other_item)} are "
                    "not of compatible classes."
                )
            if not isinstance(other_item, str) and (
                self._url is None
                or other_item._url is None
                or other_item._url.scheme != self._url.scheme
                or other_item._url.netloc != self._url.netloc
                or other_item._kwargs != self._kwargs
            ):
                raise ValueError(
                    f"{self} and {other_item} do not share the same "
                    "base URL and storage options."
                )
        output: PT = super().relative_to(*other)  # type: ignore
        output._url = self._url
        output._kwargs = self._kwargs
        return output

    def _scandir(self):
        # provided in Python3.11 but not required in fsspec glob implementation
        raise NotImplementedError

    def glob(self: PT, pattern: str) -> Generator[PT, None, None]:
        path_pattern = self.joinpath(pattern)
        for name in self._accessor.glob(self, path_pattern):
            name = self._sub_path(name)
            name = name.split(self._flavour.sep)
            yield self._make_child(name)

    def rglob(self: PT, pattern: str) -> Generator[PT, None, None]:
        path_pattern = self.joinpath("**", pattern)
        for name in self._accessor.glob(self, path_pattern):
            name = self._sub_path(name)
            name = name.split(self._flavour.sep)
            yield self._make_child(name)

    def _sub_path(self, name):
        # only want the path name with iterdir
        sp = self.path
        return re.sub(f"^({sp}|{sp[1:]})/", "", name)

    def absolute(self: PT) -> PT:
        # fsspec paths are always absolute
        return self

    def resolve(self: PT, strict: bool = False) -> PT:
        """Return a new path with '.' and '..' parts normalized."""
        _parts = self._parts

        # Do not attempt to normalize path if no parts are dots
        if ".." not in _parts and "." not in _parts:
            return self

        sep = self._flavour.sep

        resolved: list[str] = []
        resolvable_parts = _parts[1:]
        idx_max = len(resolvable_parts) - 1
        for i, part in enumerate(resolvable_parts):
            if part == "..":
                if resolved:
                    resolved.pop()
            elif part != ".":
                if i < idx_max:
                    part += sep
                resolved.append(part)

        path = "".join(resolved)
        url = self._url
        if url is not None:
            url = url._replace(path=path)
        parts = _parts[:1] + path.split(sep)

        return self._from_parsed_parts(
            self._drv,
            self._root,
            parts,
            url=url,
            **self._kwargs,
        )

    def exists(self) -> bool:
        """
        Whether this path exists.
        """
        if not getattr(self._accessor, "exists"):
            try:
                self._accessor.stat(self)
            except FileNotFoundError:
                return False
            return True
        else:
            return bool(self._accessor.exists(self))

    def is_dir(self) -> bool:
        try:
            info = self._accessor.info(self)
            if info["type"] == "directory":
                return True
        except FileNotFoundError:
            return False
        return False

    def is_file(self) -> bool:
        try:
            info = self._accessor.info(self)
            if info["type"] == "file":
                return True
        except FileNotFoundError:
            return False
        return False

    def is_mount(self) -> bool:
        return False

    def is_symlink(self) -> bool:
        try:
            info = self._accessor.info(self)
            if "islink" in info:
                return bool(info["islink"])
        except FileNotFoundError:
            return False
        return False

    def is_socket(self) -> bool:
        return False

    def is_fifo(self) -> bool:
        return False

    def is_block_device(self) -> bool:
        return False

    def is_char_device(self) -> bool:
        return False

    def is_absolute(self) -> bool:
        return True

    def unlink(self, missing_ok: bool = False) -> None:
        if not self.exists():
            if not missing_ok:
                raise FileNotFoundError
            else:
                return
        self._accessor.rm(self, recursive=False)

    def rmdir(self, recursive: bool = True) -> None:
        """Add warning if directory not empty
        assert is_dir?
        """
        if not self.is_dir():
            raise NotADirectoryError
        self._accessor.rm(self, recursive=recursive)

    def chmod(self, mode, *, follow_symlinks: bool = True) -> None:
        raise NotImplementedError

    def rename(self, target):
        # can be implemented, but may be tricky
        raise NotImplementedError

    def replace(self, target):
        raise NotImplementedError

    def symlink_to(self, target, target_is_directory=False):
        raise NotImplementedError

    def hardlink_to(self, target):
        raise NotImplementedError

    def link_to(self, target):
        raise NotImplementedError

    def cwd(self):
        raise NotImplementedError

    def home(self):
        raise NotImplementedError

    def expanduser(self):
        raise NotImplementedError

    def group(self):
        raise NotImplementedError

    def lchmod(self, mode):
        raise NotImplementedError

    def lstat(self):
        raise NotImplementedError

    def owner(self):
        raise NotImplementedError

    def readlink(self):
        raise NotImplementedError

    def touch(self, *args: int, truncate: bool = True, **kwargs) -> None:
        # Keep the calling signature compatible with Path
        # (without changing current fsspec behavior for defaults)
        if len(args) > 2:
            raise TypeError("too many arguments")
        else:
            for key, val in zip(["mode", "exists_ok"], args):
                if key in kwargs:
                    raise TypeError(f"provided {key!r} as arg and kwarg")
                kwargs[key] = val
        self._accessor.touch(self, truncate=truncate, **kwargs)

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False
    ) -> None:
        """
        Create a new directory at this given path.
        """
        if parents:
            if not exist_ok and self.exists():
                raise FileExistsError
            self._accessor.makedirs(self, exist_ok=exist_ok)
        else:
            try:
                self._accessor.mkdir(
                    self,
                    create_parents=False,
                    mode=mode,
                )
            except FileExistsError:
                if not exist_ok or not self.is_dir():
                    raise

    @classmethod
    def _from_parts(
        cls: type[PT],
        args: list[str | PathLike],
        url: SplitResult | None = None,
        **kwargs: Any,
    ) -> PT:
        obj = object.__new__(cls)
        drv, root, parts = obj._parse_args(args)
        obj._drv = drv
        if sys.version_info < (3, 9):
            obj._closed = False
        obj._url = url
        obj._kwargs = kwargs.copy()

        if not root:
            if not parts:
                root = "/"
                parts = ["/"]
            elif parts[0] == "/":
                root = parts[1:]
        obj._root = root
        obj._parts = parts

        return obj

    @classmethod
    def _from_parsed_parts(
        cls: type[PT],
        drv: str,
        root: str,
        parts: list[str],
        url: SplitResult | None = None,
        **kwargs: Any,
    ) -> PT:
        obj = object.__new__(cls)
        obj._drv = drv
        obj._parts = parts
        if sys.version_info < (3, 9):
            obj._closed = False
        obj._url = url
        obj._kwargs = kwargs.copy()

        if not root:
            if not parts:
                root = "/"
            elif parts[0] == "/":
                root = parts.pop(0)
        if len(obj._parts) == 0 or obj._parts[0] != root:
            obj._parts.insert(0, root)

        obj._root = root
        return obj

    def __str__(self) -> str:
        """Return the string representation of the path, suitable for
        passing to system calls."""
        try:
            return self._str
        except AttributeError:
            self._str = self._format_parsed_parts(
                self._drv,
                self._root,
                self._parts,
                url=self._url,
                **self._kwargs,
            )
            return self._str

    @property
    def fs(self) -> AbstractFileSystem:
        return self._accessor._fs

    def __truediv__(self: PT, key: str | PathLike) -> PT:
        # Add `/` root if not present
        if len(self._parts) == 0:
            key = f"{self._root}{key}"

        # Adapted from `PurePath._make_child`
        drv, root, parts = self._parse_args((key,))
        drv, root, parts = self._flavour.join_parsed_parts(
            self._drv, self._root, self._parts, drv, root, parts
        )

        kwargs = self._kwargs.copy()

        # Create a new object
        out = self.__class__(
            self._format_parsed_parts(drv, root, parts, url=self._url),
            **kwargs,
        )
        return out

    def __setstate__(self, state: dict) -> None:
        self._kwargs = state["_kwargs"].copy()

    def __reduce__(self):
        cls = type(self)
        return (
            cls,
            (
                cls._format_parsed_parts(
                    self._drv, self._root, self._parts, url=self._url
                ),
            ),
            {"_kwargs": self._kwargs.copy()},
        )

    def with_suffix(self: PT, suffix: str) -> PT:
        """Return a new path with the file suffix changed.  If the path
        has no suffix, add given suffix.  If the given suffix is an empty
        string, remove the suffix from the path.
        """
        f = self._flavour
        if f.sep in suffix or f.altsep and f.altsep in suffix:
            raise ValueError("Invalid suffix %r" % (suffix,))
        if suffix and not suffix.startswith(".") or suffix == ".":
            raise ValueError("Invalid suffix %r" % (suffix))
        name = self.name
        if not name:
            raise ValueError("%r has an empty name" % (self,))
        old_suffix = self.suffix
        if not old_suffix:
            name = name + suffix
        else:
            name = name[: -len(old_suffix)] + suffix
        return self._from_parsed_parts(
            self._drv,
            self._root,
            self._parts[:-1] + [name],
            url=self._url,
            **self._kwargs,
        )

    def with_name(self: PT, name: str) -> PT:
        """Return a new path with the file name changed."""
        if not self.name:
            raise ValueError("%r has an empty name" % (self,))
        drv, root, parts = self._flavour.parse_parts((name,))
        if (
            not name
            or name[-1] in [self._flavour.sep, self._flavour.altsep]
            or drv
            or root
            or len(parts) != 1
        ):
            raise ValueError("Invalid name %r" % (name))
        return self._from_parsed_parts(
            self._drv,
            self._root,
            self._parts[:-1] + [name],
            url=self._url,
            **self._kwargs,
        )

    @property
    def parents(self) -> _UPathParents:
        """A sequence of this upath's logical parents."""
        return _UPathParents(self)


class _UPathParents(Sequence[UPath]):
    """This object provides sequence-like access to the logical ancestors
    of a path.  Don't try to construct it yourself."""

    __slots__ = (
        "_pathcls",
        "_drv",
        "_root",
        "_parts",
        "_url",
        "_kwargs",
    )

    def __init__(self, path):
        # We don't store the instance to avoid reference cycles
        self._pathcls = type(path)
        self._drv = path._drv
        self._root = path._root
        self._parts = path._parts
        self._url = path._url
        self._kwargs = path._kwargs

    def __len__(self):
        if self._drv or self._root:
            return len(self._parts) - 1
        else:
            return len(self._parts)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(*idx.indices(len(self))))

        if idx >= len(self) or idx < -len(self):
            raise IndexError(idx)
        if idx < 0:
            idx += len(self)
        return self._pathcls._from_parsed_parts(
            self._drv,
            self._root,
            self._parts[: -idx - 1],
            url=self._url,
            **self._kwargs,
        )

    def __repr__(self):
        return "<{}.parents>".format(self._pathcls.__name__)



from fsspec.asyn import sync

import upath.core


class _HTTPAccessor(upath.core._FSSpecAccessor):
    def __init__(self, parsed_url, *args, **kwargs):
        super().__init__(parsed_url, *args, **kwargs)

    def _format_path(self, path):
        return str(path)


class HTTPPath(upath.core.UPath):
    _default_accessor = _HTTPAccessor

    def is_dir(self):
        try:
            return self._path_type() == "directory"
        except FileNotFoundError:
            return False

    def is_file(self):
        try:
            return self._path_type() == "file"
        except FileNotFoundError:
            return False

    def _path_type(self):
        try:
            next(self.iterdir())
        except (StopIteration, NotADirectoryError):
            return "file"
        else:
            return "directory"

    def _sub_path(self, name):
        """
        `fsspec` returns the full path as `scheme://netloc/<path>` with
        `listdir` and `glob`. However, in `iterdir` and `glob` we only want the
        relative path to `self`.
        """
        complete_address = self._format_parsed_parts(
            None, None, [self.path], url=self._url, **self._kwargs
        )

        if name.startswith(complete_address):
            name = name[len(complete_address) :]  # noqa: E203
        name = name.strip("/")

        return name

    def resolve(
        self: "HTTPPath", strict: bool = False, follow_redirects: bool = True
    ) -> "HTTPPath":
        """Normalize the path and resolve redirects."""
        # Normalise the path
        resolved_path = super().resolve(strict=strict)

        if follow_redirects:
            # Ensure we have a url
            parsed_url = resolved_path._url
            if parsed_url is None:
                return resolved_path
            else:
                url = parsed_url.geturl()
            # Get the fsspec fs
            fs = resolved_path._accessor._fs
            # Ensure we have a session
            session = sync(fs.loop, fs.set_session)
            # Use HEAD requests if the server allows it, falling back to GETs
            for method in (session.head, session.get):
                r = sync(fs.loop, method, url, allow_redirects=True)
                try:
                    r.raise_for_status()
                except Exception as exc:
                    if method == session.get:
                        raise FileNotFoundError(self) from exc
                else:
                    resolved_path = HTTPPath(str(r.url))
                    break

        return resolved_path



from __future__ import annotations

import importlib
import warnings
from functools import lru_cache
from typing import TYPE_CHECKING

from fsspec.core import get_filesystem_class

if TYPE_CHECKING:
    from upath.core import PT

__all__ = [
    "get_upath_class",
]


class _Registry:
    known_implementations: dict[str, str] = {
        "abfs": "upath.implementations.cloud.AzurePath",
        "adl": "upath.implementations.cloud.AzurePath",
        "az": "upath.implementations.cloud.AzurePath",
        "gcs": "upath.implementations.cloud.GCSPath",
        "gs": "upath.implementations.cloud.GCSPath",
        "hdfs": "upath.implementations.hdfs.HDFSPath",
        "http": "upath.implementations.http.HTTPPath",
        "https": "upath.implementations.http.HTTPPath",
        "memory": "upath.implementations.memory.MemoryPath",
        "s3": "upath.implementations.cloud.S3Path",
        "s3a": "upath.implementations.cloud.S3Path",
        "webdav+http": "upath.implementations.webdav.WebdavPath",
        "webdav+https": "upath.implementations.webdav.WebdavPath",
    }

    def __getitem__(self, item: str) -> type[PT] | None:
        try:
            fqn = self.known_implementations[item]
        except KeyError:
            return None
        module_name, name = fqn.rsplit(".", 1)
        mod = importlib.import_module(module_name)
        return getattr(mod, name)  # type: ignore


_registry = _Registry()


@lru_cache()
def get_upath_class(protocol: str) -> type[PT] | None:
    """return the upath cls for the given protocol"""
    cls: type[PT] | None = _registry[protocol]
    if cls is not None:
        return cls
    else:
        if not protocol:
            return None  # we want to use pathlib for `None` protocol
        try:
            _fs_cls = get_filesystem_class(protocol)
        except ValueError:
            return None  # this is an unknown protocol
        else:
            if _fs_cls.protocol != "file":
                warnings.warn(
                    f"UPath {protocol!r} filesystem not explicitly implemented."
                    " Falling back to default implementation."
                    " This filesystem may not be tested.",
                    UserWarning,
                    stacklevel=2,
                )
            mod = importlib.import_module("upath.core")
            return getattr(mod, "UPath")  # type: ignore



import pytest
from upath import UPath
from upath.implementations.cloud import AzurePath

from ..cases import BaseTests
from ..utils import skip_on_windows


@skip_on_windows
@pytest.mark.usefixtures("path")
class TestAzurePath(BaseTests):
    SUPPORTS_EMPTY_DIRS = False

    @pytest.fixture(autouse=True, scope="function")
    def path(self, azurite_credentials, azure_fixture):
        account_name, connection_string = azurite_credentials

        self.storage_options = {
            "account_name": account_name,
            "connection_string": connection_string,
        }
        self.path = UPath(azure_fixture, **self.storage_options)
        self.prepare_file_system()

    def test_is_AzurePath(self):
        assert isinstance(self.path, AzurePath)

    def test_rmdir(self):
        new_dir = self.path / "new_dir_rmdir"
        new_dir.mkdir()
        path = new_dir / "test.txt"
        path.write_text("hello")
        assert path.exists()
        new_dir.fs.invalidate_cache()
        new_dir.rmdir()
        assert not new_dir.exists()

        with pytest.raises(NotADirectoryError):
            (self.path / "a" / "file.txt").rmdir()

    @pytest.mark.skip
    def test_makedirs_exist_ok_false(self):
        pass



import pickle
from pathlib import Path
import re
import sys

import pytest
from upath import UPath


class BaseTests:
    SUPPORTS_EMPTY_DIRS = True

    path: UPath

    def test_cwd(self):
        with pytest.raises(NotImplementedError):
            self.path.cwd()

    def test_home(self):
        with pytest.raises(NotImplementedError):
            self.path.home()

    def test_stat(self):
        stat = self.path.stat()
        assert stat

    def test_chmod(self):
        with pytest.raises(NotImplementedError):
            self.path.joinpath("file1.txt").chmod(777)

    @pytest.mark.parametrize(
        "url, expected", [("file1.txt", True), ("fakefile.txt", False)]
    )
    def test_exists(self, url, expected):
        path = self.path.joinpath(url)
        assert path.exists() == expected

    def test_expanduser(self):
        with pytest.raises(NotImplementedError):
            self.path.expanduser()

    def test_glob(self, pathlib_base):
        mock_glob = list(self.path.glob("**.txt"))
        path_glob = list(pathlib_base.glob("**/*.txt"))

        _mock_start = len(self.path.parts)
        mock_glob_normalized = sorted(
            [a.parts[_mock_start:] for a in mock_glob]
        )
        _path_start = len(pathlib_base.parts)
        path_glob_normalized = sorted(
            [a.parts[_path_start:] for a in path_glob]
        )

        print(mock_glob_normalized, path_glob_normalized)
        assert mock_glob_normalized == path_glob_normalized

    def test_group(self):
        with pytest.raises(NotImplementedError):
            self.path.group()

    def test_is_dir(self):
        assert self.path.is_dir()

        path = self.path / "file1.txt"
        assert not path.is_dir()
        assert not (self.path / "not-existing-dir").is_dir()

    def test_is_file(self):
        path = self.path / "file1.txt"
        assert path.is_file()
        assert not self.path.is_file()

        assert not (self.path / "not-existing-file.txt").is_file()

    def test_is_mount(self):
        assert self.path.is_mount() is False

    def test_is_symlink(self):
        assert self.path.is_symlink() is False

    def test_is_socket(self):
        assert self.path.is_socket() is False

    def test_is_fifo(self):
        assert self.path.is_fifo() is False

    def test_is_block_device(self):
        assert self.path.is_block_device() is False

    def test_is_char_device(self):
        assert self.path.is_char_device() is False

    def test_iterdir(self, local_testdir):
        pl_path = Path(local_testdir)

        up_iter = list(self.path.iterdir())
        pl_iter = list(pl_path.iterdir())

        for x in up_iter:
            assert x.name != ""
            assert x.exists()

        assert len(up_iter) == len(pl_iter)
        assert set(p.name for p in pl_iter) == set(u.name for u in up_iter)
        assert next(self.path.parent.iterdir()).exists()

    def test_iterdir2(self, local_testdir):
        pl_path = Path(local_testdir) / "folder1"

        up_iter = list((self.path / "folder1").iterdir())
        pl_iter = list(pl_path.iterdir())

        for x in up_iter:
            assert x.exists()

        assert len(up_iter) == len(pl_iter)
        assert set(p.name for p in pl_iter) == set(u.name for u in up_iter)
        assert next(self.path.parent.iterdir()).exists()

    def test_parents(self):
        p = self.path.joinpath("folder1", "file1.txt")
        assert p.is_file()
        assert p.parents[0] == p.parent
        assert p.parents[1] == p.parent.parent
        assert p.parents[0].name == "folder1"
        assert p.parents[1].name == self.path.name

    def test_lchmod(self):
        with pytest.raises(NotImplementedError):
            self.path.lchmod(mode=77)

    def test_lstat(self):
        with pytest.raises(NotImplementedError):
            self.path.lstat()

    def test_mkdir(self):
        new_dir = self.path.joinpath("new_dir")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        assert new_dir.exists()

    def test_mkdir_exists_ok_true(self):
        new_dir = self.path.joinpath("new_dir_may_exists")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        new_dir.mkdir(exist_ok=True)

    def test_mkdir_exists_ok_false(self):
        new_dir = self.path.joinpath("new_dir_may_not_exists")
        new_dir.mkdir()
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        with pytest.raises(FileExistsError):
            new_dir.mkdir(exist_ok=False)

    def test_mkdir_parents_true_exists_ok_true(self):
        new_dir = self.path.joinpath("parent", "new_dir_may_not_exist")
        new_dir.mkdir(parents=True)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        new_dir.mkdir(parents=True, exist_ok=True)

    def test_mkdir_parents_true_exists_ok_false(self):
        new_dir = self.path.joinpath("parent", "new_dir_may_exist")
        new_dir.mkdir(parents=True)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        with pytest.raises(FileExistsError):
            new_dir.mkdir(parents=True, exist_ok=False)

    def test_makedirs_exist_ok_true(self):
        new_dir = self.path.joinpath("parent", "child", "dir_may_not_exist")
        new_dir._accessor.makedirs(new_dir, exist_ok=True)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        new_dir._accessor.makedirs(new_dir, exist_ok=True)

    def test_makedirs_exist_ok_false(self):
        new_dir = self.path.joinpath("parent", "child", "dir_may_exist")
        new_dir._accessor.makedirs(new_dir, exist_ok=False)
        if not self.SUPPORTS_EMPTY_DIRS:
            new_dir.joinpath(".file").touch()
        with pytest.raises(FileExistsError):
            new_dir._accessor.makedirs(new_dir, exist_ok=False)

    def test_open(self):
        pass

    def test_owner(self):
        with pytest.raises(NotImplementedError):
            self.path.owner()

    def test_read_bytes(self, pathlib_base):
        mock = self.path.joinpath("file2.txt")
        pl = pathlib_base.joinpath("file2.txt")
        assert mock.read_bytes() == pl.read_bytes()

    def test_read_text(self, local_testdir):
        upath = self.path.joinpath("file1.txt")
        assert (
            upath.read_text()
            == Path(local_testdir).joinpath("file1.txt").read_text()
        )

    def test_readlink(self):
        with pytest.raises(NotImplementedError):
            self.path.readlink()

    @pytest.mark.xfail
    def test_rename(self):
        # need to implement
        raise False

    def test_replace(self):
        pass

    def test_resolve(self):
        pass

    def test_rglob(self):
        pass

    def test_samefile(self):
        pass

    def test_symlink_to(self):
        pass

    def test_touch_unlink(self):
        path = self.path.joinpath("test_touch.txt")
        path.touch()
        assert path.exists()
        path.unlink()
        assert not path.exists()

        # should raise FileNotFoundError since file is missing
        with pytest.raises(FileNotFoundError):
            path.unlink()

        # file doesn't exists, but missing_ok is True
        path.unlink(missing_ok=True)

    def test_link_to(self):
        pass

    def test_write_bytes(self, pathlib_base):
        fn = "test_write_bytes.txt"
        s = b"hello_world"
        path = self.path.joinpath(fn)
        path.write_bytes(s)
        assert path.read_bytes() == s

    def test_write_text(self, pathlib_base):
        fn = "test_write_text.txt"
        s = "hello_world"
        path = self.path.joinpath(fn)
        path.write_text(s)
        assert path.read_text() == s

    def prepare_file_system(self):
        self.make_top_folder()
        self.make_test_files()

    def make_top_folder(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def make_test_files(self):
        folder1 = self.path.joinpath("folder1")
        folder1.mkdir(exist_ok=True)
        folder1_files = ["file1.txt", "file2.txt"]
        for f in folder1_files:
            p = folder1.joinpath(f)
            p.touch()
            p.write_text(f)

        file1 = self.path.joinpath("file1.txt")
        file1.touch()
        file1.write_text("hello world")
        file2 = self.path.joinpath("file2.txt")
        file2.touch()
        file2.write_bytes(b"hello world")

    def test_fsspec_compat(self):
        fs = self.path.fs
        content = b"a,b,c\n1,2,3\n4,5,6"

        def strip_scheme(path):
            root = "" if sys.platform.startswith("win") else "/"
            return root + re.sub("^[a-z0-9]+:/*", "", str(path))

        upath1 = self.path / "output1.csv"
        p1 = strip_scheme(upath1)
        upath1.write_bytes(content)
        assert fs is upath1.fs
        with fs.open(p1) as f:
            assert f.read() == content
        upath1.unlink()

        # write with fsspec, read with upath
        upath2 = self.path / "output2.csv"
        p2 = strip_scheme(upath2)
        assert fs is upath2.fs
        with fs.open(p2, "wb") as f:
            f.write(content)
        assert upath2.read_bytes() == content
        upath2.unlink()

    def test_pickling(self):
        path = self.path
        pickled_path = pickle.dumps(path)
        recovered_path = pickle.loads(pickled_path)

        assert type(path) == type(recovered_path)
        assert str(path) == str(recovered_path)
        assert path.fs.storage_options == recovered_path.fs.storage_options

    def test_pickling_child_path(self):
        path = self.path / "subfolder" / "subsubfolder"
        pickled_path = pickle.dumps(path)
        recovered_path = pickle.loads(pickled_path)

        assert type(path) == type(recovered_path)
        assert str(path) == str(recovered_path)
        assert path._drv == recovered_path._drv
        assert path._root == recovered_path._root
        assert path._parts == recovered_path._parts
        assert path.fs.storage_options == recovered_path.fs.storage_options

    def test_child_path(self):
        path_str = str(self.path).rstrip("/")
        path_a = UPath(f"{path_str}/folder")
        path_b = self.path / "folder"

        assert str(path_a) == str(path_b)
        assert path_a._root == path_b._root
        assert path_a._drv == path_b._drv
        assert path_a._parts == path_b._parts
        assert path_a._url == path_b._url

    def test_copy_path(self):
        path = self.path
        copy_path = UPath(path)

        assert type(path) == type(copy_path)
        assert str(path) == str(copy_path)
        assert path._drv == copy_path._drv
        assert path._root == copy_path._root
        assert path._parts == copy_path._parts
        assert path.fs.storage_options == copy_path.fs.storage_options

    def test_with_name(self):
        path = self.path / "file.txt"
        path = path.with_name("file.zip")
        assert path.name == "file.zip"

    def test_with_suffix(self):
        path = self.path / "file.txt"
        path = path.with_suffix(".zip")
        assert path.suffix == ".zip"

    def test_with_stem(self):
        if sys.version_info < (3, 9):
            pytest.skip("with_stem only available on py3.9+")
        path = self.path / "file.txt"
        path = path.with_stem("document")
        assert path.stem == "document"

    def test_repr_after_with_name(self):
        p = self.path.joinpath("file.txt").with_name("file.zip")
        assert "file.zip" in repr(p)

    def test_repr_after_with_suffix(self):
        p = self.path.joinpath("file.txt").with_suffix(".zip")
        assert "file.zip" in repr(p)

    def test_rmdir_no_dir(self):
        p = self.path.joinpath("file1.txt")
        with pytest.raises(NotADirectoryError):
            p.rmdir()

    def test_iterdir_no_dir(self):
        p = self.path.joinpath("file1.txt")
        assert p.is_file()
        with pytest.raises(NotADirectoryError):
            _ = list(p.iterdir())



import pytest

from upath import UPath
from upath.implementations.cloud import GCSPath
from ..cases import BaseTests
from ..utils import skip_on_windows


@skip_on_windows
@pytest.mark.usefixtures("path")
class TestGCSPath(BaseTests):
    SUPPORTS_EMPTY_DIRS = False

    @pytest.fixture(autouse=True, scope="function")
    def path(self, gcs_fixture):
        path, endpoint_url = gcs_fixture
        self.path = UPath(path, endpoint_url=endpoint_url)
        self.endpoint_url = endpoint_url

    def test_is_GCSPath(self):
        assert isinstance(self.path, GCSPath)

    def test_rmdir(self):
        dirname = "rmdir_test"
        mock_dir = self.path.joinpath(dirname)
        mock_dir.joinpath("test.txt").write_text("hello")
        mock_dir.fs.invalidate_cache()
        mock_dir.rmdir()
        assert not mock_dir.exists()
        with pytest.raises(NotADirectoryError):
            self.path.joinpath("file1.txt").rmdir()

    @pytest.mark.skip
    def test_makedirs_exist_ok_false(self):
        pass



import upath.core


class _MemoryAccessor(upath.core._FSSpecAccessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fs.root_marker = ""


class MemoryPath(upath.core.UPath):
    _default_accessor = _MemoryAccessor

    def iterdir(self):
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for name in self._accessor.listdir(self):
            # fsspec returns dictionaries
            if isinstance(name, dict):
                name = name.get("name")
            if name in {".", ".."}:
                # Yielding a path object for these makes little sense
                continue
            # only want the path name with iterdir
            name = name.rstrip("/")
            name = self._sub_path(name)
            yield self._make_child_relpath(name)



"""see upath/tests/conftest.py for fixtures
"""
import pytest  # noqa: F401

from upath import UPath
from upath.implementations.hdfs import HDFSPath
from ..cases import BaseTests


@pytest.mark.hdfs
class TestUPathHDFS(BaseTests):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir, hdfs):
        host, user, port = hdfs
        path = f"hdfs:{local_testdir}"
        self.path = UPath(path, host=host, user=user, port=port)

    def test_is_HDFSPath(self):
        assert isinstance(self.path, HDFSPath)

    def test_chmod(self):
        # todo
        pass

    def test_fsspec_compat(self):
        pass



import upath.core


class _HDFSAccessor(upath.core._FSSpecAccessor):
    def __init__(self, parsed_url, *args, **kwargs):
        super().__init__(parsed_url, *args, **kwargs)
        self._fs.root_marker = "/"

    def touch(self, path, **kwargs):
        kwargs.pop("truncate", None)
        super().touch(path, **kwargs)

    def mkdir(self, path, create_parents=True, **kwargs):
        pth = self._format_path(path)
        if create_parents:
            return self._fs.makedirs(pth, **kwargs)
        else:
            if not kwargs.get("exist_ok", False) and self._fs.exists(pth):
                raise FileExistsError
            return self._fs.mkdir(pth, create_parents=create_parents, **kwargs)


class HDFSPath(upath.core.UPath):
    _default_accessor = _HDFSAccessor



import pytest  # noqa: F401

from upath import UPath
from ..cases import BaseTests


class TestUPathWebdav(BaseTests):
    @pytest.fixture(autouse=True, scope="function")
    def path(self, webdav_fixture):
        self.path = UPath(webdav_fixture, auth=("USER", "PASSWORD"))

    def test_fsspec_compat(self):
        pass



import pytest  # noqa: F401

from fsspec import get_filesystem_class

from upath import UPath
from upath.implementations.http import HTTPPath
from ..cases import BaseTests
from ..utils import skip_on_windows

try:
    get_filesystem_class("http")
except ImportError:
    pytestmark = pytest.mark.skip


def test_httppath():
    path = UPath("http://example.com")
    assert isinstance(path, HTTPPath)
    assert path.exists()


def test_httpspath():
    path = UPath("https://example.com")
    assert isinstance(path, HTTPPath)
    assert path.exists()


@skip_on_windows
class TestUPathHttp(BaseTests):
    @pytest.fixture(autouse=True, scope="function")
    def path(self, http_fixture):
        self.path = UPath(http_fixture)

    def test_work_at_root(self):
        assert "folder" in (f.name for f in self.path.parent.iterdir())

    @pytest.mark.skip
    def test_mkdir(self):
        pass

    @pytest.mark.skip
    def test_mkdir_exists_ok_false(self):
        pass

    @pytest.mark.skip
    def test_mkdir_exists_ok_true(self):
        pass

    @pytest.mark.skip
    def test_mkdir_parents_true_exists_ok_true(self):
        pass

    @pytest.mark.skip
    def test_mkdir_parents_true_exists_ok_false(self):
        pass

    @pytest.mark.skip
    def test_makedirs_exist_ok_true(self):
        pass

    @pytest.mark.skip
    def test_makedirs_exist_ok_false(self):
        pass

    @pytest.mark.skip
    def test_touch_unlink(self):
        pass

    @pytest.mark.skip
    def test_write_bytes(self, pathlib_base):
        pass

    @pytest.mark.skip
    def test_write_text(self, pathlib_base):
        pass

    def test_fsspec_compat(self):
        pass

    def test_resolve(self):
        # Also tests following redirects, because the test server issues a
        # 301 redirect for `http://127.0.0.1:8080/folder` to
        # `http://127.0.0.1:8080/folder/`
        assert str(self.path.resolve()).endswith("/")



from urllib.parse import ParseResult
from urllib.parse import urlunsplit

import upath.core


class _WebdavAccessor(upath.core._FSSpecAccessor):
    def __init__(self, parsed_url: ParseResult, **kwargs):
        from webdav4.fsspec import WebdavFileSystem

        parsed_url = parsed_url._replace(scheme=parsed_url.scheme[7:], path="")
        base_url = urlunsplit(parsed_url)
        self._fs = WebdavFileSystem(base_url, **kwargs)

    def listdir(self, path, **kwargs):
        base_url = urlunsplit(path._url._replace(path=""))
        for file_info in self._fs.listdir(
            self._format_path(path).lstrip("/"), **kwargs
        ):
            yield {
                **file_info,
                "name": f"{base_url}/{file_info['name']}",
            }

    def glob(self, path, path_pattern, **kwargs):
        base_url = urlunsplit(path._url._replace(path=""))
        for file_path in self._fs.glob(
            self._format_path(path_pattern).lstrip("/"), **kwargs
        ):
            yield f"{base_url}/{file_path}"


class WebdavPath(upath.core.UPath):
    _default_accessor = _WebdavAccessor

    def _sub_path(self, name):
        """fsspec returns path as `scheme://netloc/<path>` with listdir
        and glob, so we potentially need to sub the whole string
        """
        sp = self.path
        complete_address = self._format_parsed_parts(
            None, None, [sp], url=self._url, **self._kwargs
        )

        if name.startswith(complete_address):
            name = name[len(complete_address) :]  # noqa: E203
        name = name.strip("/")

        return name



import pathlib
import pickle
import sys
import warnings

import pytest
from upath import UPath
from upath.implementations.cloud import S3Path
from .cases import BaseTests
from .utils import only_on_windows, skip_on_windows


@skip_on_windows
def test_posix_path(local_testdir):
    assert isinstance(UPath(local_testdir), pathlib.PosixPath)


@only_on_windows
def test_windows_path(local_testdir):
    assert isinstance(UPath(local_testdir), pathlib.WindowsPath)


def test_UPath_untested_protocol_warning():
    with warnings.catch_warnings(record=True) as w:
        _ = UPath("mock:/")
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "mock" in str(w[-1].message)


def test_UPath_file_protocol_no_warning():
    with warnings.catch_warnings(record=True) as w:
        _ = UPath("file:/")
        assert len(w) == 0


class TestUpath(BaseTests):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # On Windows the path needs to be prefixed with `/`, becaue
            # `UPath` implements `_posix_flavour`, which requires a `/` root
            # in order to correctly deserialize pickled objects
            root = "/" if sys.platform.startswith("win") else ""
            self.path = UPath(f"mock:{root}{local_testdir}")

    def test_fsspec_compat(self):
        pass


@pytest.mark.hdfs
def test_multiple_backend_paths(local_testdir, s3_fixture, hdfs):
    _, anon, s3so = s3_fixture
    path = f"s3:{local_testdir}"
    s3_path = UPath(path, anon=anon, **s3so)
    assert s3_path.joinpath("text.txt")._url.scheme == "s3"
    host, user, port = hdfs
    path = f"hdfs:{local_testdir}"
    UPath(path, host=host, user=user, port=port)
    assert s3_path.joinpath("text1.txt")._url.scheme == "s3"


def test_constructor_accept_path(local_testdir):
    path = UPath(pathlib.Path(local_testdir))
    assert str(path) == str(pathlib.Path(local_testdir))


def test_constructor_accept_upath(local_testdir):
    path = UPath(UPath(local_testdir))
    assert str(path) == str(pathlib.Path(local_testdir))


def test_subclass(local_testdir):
    class MyPath(UPath):
        pass

    path = MyPath(local_testdir)
    assert str(path) == str(pathlib.Path(local_testdir))
    assert issubclass(MyPath, UPath)
    assert isinstance(path, pathlib.Path)


def test_subclass_with_gcs():
    path = UPath("gcs://bucket", anon=True)
    assert isinstance(path, UPath)
    assert isinstance(path, pathlib.Path)


def test_instance_check(local_testdir):
    path = pathlib.Path(local_testdir)
    upath = UPath(local_testdir)
    # test instance check passes
    assert isinstance(upath, pathlib.Path)
    assert not isinstance(upath, UPath)
    # test type is same as pathlib
    assert type(upath) is type(path)
    upath = UPath(f"file://{local_testdir}")
    # test default implementation is used
    assert type(upath) is UPath


def test_new_method(local_testdir):
    path = UPath.__new__(pathlib.Path, local_testdir)
    assert str(path) == str(pathlib.Path(local_testdir))
    assert isinstance(path, pathlib.Path)
    assert not isinstance(path, UPath)


@skip_on_windows
class TestFSSpecLocal(BaseTests):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        path = f"file://{local_testdir}"
        self.path = UPath(path)


PATHS = (
    ("path", "storage_options", "module", "object_type"),
    (
        ("/tmp/abc", (), None, pathlib.Path),
        ("s3://bucket/folder", ({"anon": True}), "s3fs", S3Path),
        ("gs://bucket/folder", ({"token": "anon"}), "gcsfs", UPath),
    ),
)


@pytest.mark.parametrize(*PATHS)
def test_create_from_type(path, storage_options, module, object_type):
    """Test that derived paths use same fs instance."""
    if module:
        # skip if module cannot be imported
        pytest.importorskip(module)
    try:
        upath = UPath(path, storage_options=storage_options)
        # test expected object type
        assert isinstance(upath, object_type)
        cast = type(upath)
        parent = upath.parent
        # test derived object is same type
        assert isinstance(parent, cast)
        # test that created fs uses fsspec instance cache
        assert not hasattr(upath, "fs") or upath.fs is parent.fs
        new = cast(str(parent))
        # test that object cast is same type
        assert isinstance(new, cast)
    except (ImportError, ModuleNotFoundError):
        # fs failed to import
        pass


def test_child_path():
    path_a = UPath("gcs://bucket/folder")
    path_b = UPath("gcs://bucket") / "folder"

    assert str(path_a) == str(path_b)
    assert path_a._root == path_b._root
    assert path_a._drv == path_b._drv
    assert path_a._parts == path_b._parts
    assert path_a._url == path_b._url


def test_pickling():
    path = UPath("gcs://bucket/folder", storage_options={"anon": True})
    pickled_path = pickle.dumps(path)
    recovered_path = pickle.loads(pickled_path)

    assert type(path) == type(recovered_path)
    assert str(path) == str(recovered_path)
    assert path.fs.storage_options == recovered_path.fs.storage_options


def test_pickling_child_path():
    path = UPath("gcs://bucket", anon=True) / "subfolder" / "subsubfolder"
    pickled_path = pickle.dumps(path)
    recovered_path = pickle.loads(pickled_path)

    assert type(path) == type(recovered_path)
    assert str(path) == str(recovered_path)
    assert path._drv == recovered_path._drv
    assert path._root == recovered_path._root
    assert path._parts == recovered_path._parts
    assert path.fs.storage_options == recovered_path.fs.storage_options


def test_copy_path():
    path = UPath("gcs://bucket/folder", anon=True)
    copy_path = UPath(path)

    assert type(path) == type(copy_path)
    assert str(path) == str(copy_path)
    assert path._drv == copy_path._drv
    assert path._root == copy_path._root
    assert path._parts == copy_path._parts
    assert path.fs.storage_options == copy_path.fs.storage_options


def test_copy_path_posix():
    path = UPath("/tmp/folder")
    copy_path = UPath(path)

    assert type(path) == type(copy_path) == type(pathlib.Path(""))
    assert str(path) == str(copy_path)
    assert path._drv == copy_path._drv
    assert path._root == copy_path._root
    assert path._parts == copy_path._parts


def test_copy_path_append():
    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2")

    assert type(path) == type(copy_path) == type(pathlib.Path(""))
    assert str(path / "folder2") == str(copy_path)

    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2/folder3")

    assert str(path / "folder2" / "folder3") == str(copy_path)

    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2", "folder3")

    assert str(path / "folder2" / "folder3") == str(copy_path)


def test_copy_path_append_kwargs():
    path = UPath("gcs://bucket/folder", anon=True)
    copy_path = UPath(path, anon=False)

    assert type(path) == type(copy_path)
    assert str(path) == str(copy_path)
    assert not copy_path._kwargs["anon"]
    assert path._kwargs["anon"]


def test_relative_to():
    assert "s3://test_bucket/file.txt" == str(
        UPath("s3://test_bucket/file.txt").relative_to(
            UPath("s3://test_bucket")
        )
    )

    with pytest.raises(ValueError):
        UPath("s3://test_bucket/file.txt").relative_to(
            UPath("gcs://test_bucket")
        )

    with pytest.raises(ValueError):
        UPath("s3://test_bucket/file.txt", anon=True).relative_to(
            UPath("s3://test_bucket", anon=False)
        )


def test_uri_parsing():
    assert (
        str(UPath("http://www.example.com//a//b/"))
        == "http://www.example.com//a//b/"
    )


NORMALIZATIONS = (
    ("unnormalized", "normalized"),
    (
        # Expected normalization results according to curl
        ("http://example.com", "http://example.com/"),
        ("http://example.com/", "http://example.com/"),
        ("http://example.com/a", "http://example.com/a"),
        ("http://example.com//a", "http://example.com//a"),
        ("http://example.com///a", "http://example.com///a"),
        ("http://example.com////a", "http://example.com////a"),
        ("http://example.com/a/.", "http://example.com/a/"),
        ("http://example.com/a/./", "http://example.com/a/"),
        ("http://example.com/a/./b", "http://example.com/a/b"),
        ("http://example.com/a/.//", "http://example.com/a//"),
        ("http://example.com/a/.//b", "http://example.com/a//b"),
        ("http://example.com/a//.", "http://example.com/a//"),
        ("http://example.com/a//./", "http://example.com/a//"),
        ("http://example.com/a//./b", "http://example.com/a//b"),
        ("http://example.com/a//.//", "http://example.com/a///"),
        ("http://example.com/a//.//b", "http://example.com/a///b"),
        ("http://example.com/a/..", "http://example.com/"),
        ("http://example.com/a/../", "http://example.com/"),
        ("http://example.com/a/../.", "http://example.com/"),
        ("http://example.com/a/../..", "http://example.com/"),
        ("http://example.com/a/../../", "http://example.com/"),
        ("http://example.com/a/../..//", "http://example.com//"),
        ("http://example.com/a/..//", "http://example.com//"),
        ("http://example.com/a/..//.", "http://example.com//"),
        ("http://example.com/a/..//..", "http://example.com/"),
        ("http://example.com/a/../b", "http://example.com/b"),
        ("http://example.com/a/..//b", "http://example.com//b"),
        ("http://example.com/a//..", "http://example.com/a/"),
        ("http://example.com/a//../", "http://example.com/a/"),
        ("http://example.com/a//../.", "http://example.com/a/"),
        ("http://example.com/a//../..", "http://example.com/"),
        ("http://example.com/a//../../", "http://example.com/"),
        ("http://example.com/a//../..//", "http://example.com//"),
        ("http://example.com/a//..//..", "http://example.com/a/"),
        ("http://example.com/a//../b", "http://example.com/a/b"),
        ("http://example.com/a//..//", "http://example.com/a//"),
        ("http://example.com/a//..//.", "http://example.com/a//"),
        ("http://example.com/a//..//b", "http://example.com/a//b"),
        # Normalization with and without an authority component
        ("memory:/a/b/..", "memory:/a/"),
        ("memory:/a/b/../..", "memory:/"),
        ("memory:/a/b/../../..", "memory:/"),
        ("memory://a/b/..", "memory://a/"),
        ("memory://a/b/../..", "memory://a/"),
        ("memory://a/b/../../..", "memory://a/"),
    ),
)


@pytest.mark.parametrize(*NORMALIZATIONS)
def test_normalize(unnormalized, normalized):
    expected = str(UPath(normalized))
    # Normalise only, do not attempt to follow redirects for http:// paths here
    result = str(UPath.resolve(UPath(unnormalized)))
    assert expected == result



import os
import shutil
import tempfile
from pathlib import Path
import subprocess
import shlex
import threading
import time
import sys

import pytest
from fsspec.implementations.local import LocalFileSystem
from fsspec.registry import register_implementation, _registry

import fsspec

from .utils import posixify


def pytest_addoption(parser):
    parser.addoption(
        "--skiphdfs", action="store_true", default=False, help="skip hdfs tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "hdfs: mark test as hdfs")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--skiphdfs"):
        return
    skip_hdfs = pytest.mark.skip(reason="skipping hdfs")
    for item in items:
        if "hdfs" in item.keywords:
            item.add_marker(skip_hdfs)


class DummyTestFS(LocalFileSystem):
    protocol = "mock"
    root_marker = "/"


@pytest.fixture(scope="session")
def clear_registry():
    register_implementation("mock", DummyTestFS)
    try:
        yield
    finally:
        _registry.clear()


@pytest.fixture(scope="function")
def tempdir(clear_registry):
    with tempfile.TemporaryDirectory() as tempdir:
        yield tempdir


@pytest.fixture(scope="function")
def local_testdir(tempdir, clear_registry):
    tmp = Path(tempdir)
    tmp.mkdir(exist_ok=True)
    folder1 = tmp.joinpath("folder1")
    folder1.mkdir()
    folder1_files = ["file1.txt", "file2.txt"]
    for f in folder1_files:
        p = folder1.joinpath(f)
        p.touch()
        p.write_text(f)

    file1 = tmp.joinpath("file1.txt")
    file1.touch()
    file1.write_text("hello world")
    file2 = tmp.joinpath("file2.txt")
    file2.touch()
    file2.write_bytes(b"hello world")
    if sys.platform.startswith("win"):
        yield str(Path(tempdir)).replace("\\", "/")
    else:
        yield tempdir


@pytest.fixture()
def pathlib_base(local_testdir):
    return Path(local_testdir)


@pytest.fixture(scope="session")
def htcluster():
    try:
        proc = subprocess.Popen(
            shlex.split("htcluster startup"),
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
    except FileNotFoundError as err:
        if err.errno == 2 and "htcluster" == err.filename:
            pytest.skip("htcluster not installed")
        raise

    time.sleep(30)
    try:
        yield
    finally:
        proc.terminate()
        proc.wait()
        proc1 = subprocess.Popen(
            shlex.split("htcluster shutdown"),
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
        proc1.terminate()
        proc1.wait()
        time.sleep(10)


@pytest.fixture()
def hdfs(htcluster, tempdir, local_testdir):
    pyarrow = pytest.importorskip("pyarrow")
    host, user, port = "0.0.0.0", "hdfs", 9000
    hdfs = pyarrow.hdfs.connect(host="0.0.0.0", port=9000, user=user)
    hdfs.mkdir(tempdir, create_parents=True)
    for x in Path(local_testdir).glob("**/*"):
        if x.is_file():
            text = x.read_text().encode("utf8")
            if not hdfs.exists(str(x.parent)):
                hdfs.mkdir(str(x.parent), create_parents=True)
            with hdfs.open(str(x), "wb") as f:
                f.write(text)
        else:
            hdfs.mkdir(str(x))
    hdfs.close()
    yield host, user, port


@pytest.fixture(scope="session")
def s3_server():
    # writable local S3 system
    if "BOTO_CONFIG" not in os.environ:  # pragma: no cover
        os.environ["BOTO_CONFIG"] = "/dev/null"
    if "AWS_ACCESS_KEY_ID" not in os.environ:  # pragma: no cover
        os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    if "AWS_SECRET_ACCESS_KEY" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    if "AWS_SECURITY_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECURITY_TOKEN"] = "testing"
    if "AWS_SESSION_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SESSION_TOKEN"] = "testing"
    if "AWS_DEFAULT_REGION" not in os.environ:  # pragma: no cover
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    requests = pytest.importorskip("requests")

    pytest.importorskip("moto")

    port = 5555
    endpoint_uri = f"http://127.0.0.1:{port}/"
    proc = subprocess.Popen(
        shlex.split(f"moto_server s3 -p {port}"),
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    try:
        timeout = 5
        while timeout > 0:
            try:
                r = requests.get(endpoint_uri, timeout=10)
                if r.ok:
                    break
            except Exception:  # pragma: no cover
                pass
            timeout -= 0.1  # pragma: no cover
            time.sleep(0.1)  # pragma: no cover
        anon = False
        s3so = dict(
            client_kwargs={"endpoint_url": endpoint_uri},
            use_listings_cache=True,
        )
        yield anon, s3so
    finally:
        proc.terminate()
        proc.wait()


@pytest.fixture
def s3_fixture(s3_server, local_testdir):
    pytest.importorskip("s3fs")
    anon, s3so = s3_server
    s3 = fsspec.filesystem("s3", anon=False, **s3so)
    bucket_name = "test_bucket"
    if s3.exists(bucket_name):
        for dir, _, keys in s3.walk(bucket_name):
            for key in keys:
                s3.rm(f"{dir}/{key}")
    else:
        s3.mkdir(bucket_name)
    for x in Path(local_testdir).glob("**/*"):
        target_path = f"{bucket_name}/{posixify(x.relative_to(local_testdir))}"
        if x.is_file():
            s3.upload(str(x), target_path)
    s3.invalidate_cache()
    yield f"s3://{bucket_name}", anon, s3so


def stop_docker(container):
    cmd = shlex.split('docker ps -a -q --filter "name=%s"' % container)
    cid = subprocess.check_output(cmd).strip().decode()
    if cid:
        subprocess.call(["docker", "rm", "-f", "-v", cid])


@pytest.fixture(scope="session")
def docker_gcs():
    if "STORAGE_EMULATOR_HOST" in os.environ:
        # assume using real API or otherwise have a server already set up
        yield os.environ["STORAGE_EMULATOR_HOST"]
        return

    requests = pytest.importorskip("requests")
    if shutil.which("docker") is None:
        pytest.skip("docker not installed")

    container = "gcsfs_test"
    cmd = (
        "docker run -d -p 4443:4443 --name gcsfs_test fsouza/fake-gcs-server:latest -scheme "  # noqa: E501
        "http -public-host http://localhost:4443 -external-url http://localhost:4443"  # noqa: E501
    )
    stop_docker(container)
    subprocess.check_output(shlex.split(cmd))
    url = "http://0.0.0.0:4443"
    timeout = 10
    while True:
        try:
            r = requests.get(url + "/storage/v1/b", timeout=10)
            if r.ok:
                yield url
                break
        except Exception as e:  # noqa: E722
            timeout -= 1
            if timeout < 0:
                raise SystemError from e
            time.sleep(1)
    stop_docker(container)


@pytest.fixture
def gcs_fixture(docker_gcs, local_testdir):
    pytest.importorskip("gcsfs")
    gcs = fsspec.filesystem("gcs", endpoint_url=docker_gcs)
    bucket_name = "test_bucket"
    if gcs.exists(bucket_name):
        for dir, _, keys in gcs.walk(bucket_name):
            for key in keys:
                gcs.rm(f"{dir}/{key}")
    else:
        gcs.mkdir(bucket_name)
    for x in Path(local_testdir).glob("**/*"):
        target_path = f"{bucket_name}/{posixify(x.relative_to(local_testdir))}"
        if x.is_file():
            gcs.upload(str(x), target_path)
    gcs.invalidate_cache()
    yield f"gs://{bucket_name}", docker_gcs


@pytest.fixture(scope="session")
def http_server():
    with tempfile.TemporaryDirectory() as tempdir:
        requests = pytest.importorskip("requests")
        pytest.importorskip("http.server")
        proc = subprocess.Popen(
            shlex.split(f"python -m http.server --directory {tempdir} 8080")
        )
        try:
            url = "http://127.0.0.1:8080/folder"
            path = Path(tempdir) / "folder"
            path.mkdir()
            timeout = 10
            while True:
                try:
                    r = requests.get(url, timeout=10)
                    if r.ok:
                        yield path, url
                        break
                except Exception as e:  # noqa: E722
                    timeout -= 1
                    if timeout < 0:
                        raise SystemError from e
                    time.sleep(1)
        finally:
            proc.terminate()
            proc.wait()


@pytest.fixture
def http_fixture(local_testdir, http_server):
    http_path, http_url = http_server
    shutil.rmtree(http_path)
    shutil.copytree(local_testdir, http_path)
    yield http_url


@pytest.fixture(scope="session")
def webdav_server(tmp_path_factory):
    try:
        from wsgidav.wsgidav_app import WsgiDAVApp
        from cheroot import wsgi
    except ImportError as err:
        pytest.skip(str(err))

    tempdir = str(tmp_path_factory.mktemp("webdav"))

    host = "127.0.0.1"
    port = 8090
    app = WsgiDAVApp(
        {
            "host": host,
            "port": port,
            "provider_mapping": {"/": tempdir},
            "simple_dc": {
                "user_mapping": {"*": {"USER": {"password": "PASSWORD"}}}
            },
        }
    )
    srvr = wsgi.Server(bind_addr=(host, port), wsgi_app=app)
    srvr.prepare()
    thread = threading.Thread(target=srvr.serve, daemon=True)
    thread.start()

    try:
        yield tempdir, f"webdav+http://{host}:{port}"
    finally:
        srvr.stop()
        thread.join()


@pytest.fixture
def webdav_fixture(local_testdir, webdav_server):
    webdav_path, webdav_url = webdav_server
    if os.path.isdir(webdav_path):
        os.rmdir(webdav_path)
    try:
        shutil.copytree(local_testdir, webdav_path)
        yield webdav_url
    finally:
        shutil.rmtree(webdav_path, ignore_errors=True)


@pytest.fixture(scope="session")
def azurite_credentials():
    url = "http://localhost:10000"
    account_name = "devstoreaccount1"
    key = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="  # noqa: E501
    endpoint = f"{url}/{account_name}"
    connection_string = f"DefaultEndpointsProtocol=http;AccountName={account_name};AccountKey={key};BlobEndpoint={endpoint};"  # noqa

    yield account_name, connection_string


@pytest.fixture(scope="session")
def docker_azurite(azurite_credentials):
    requests = pytest.importorskip("requests")

    image = "mcr.microsoft.com/azure-storage/azurite"
    container_name = "azure_test"
    cmd = (
        f"docker run --rm -d -p 10000:10000 --name {container_name} {image}"  # noqa: E501
        " azurite-blob --loose --blobHost 0.0.0.0"  # noqa: E501
    )
    url = "http://localhost:10000"

    stop_docker(container_name)
    subprocess.run(shlex.split(cmd), check=True)

    retries = 10
    while True:
        try:
            # wait until the container is up, even a 400 status code is ok
            r = requests.get(url, timeout=10)
            if (
                r.status_code == 400
                and "Server" in r.headers
                and "Azurite" in r.headers["Server"]
            ):
                yield url
                break
        except Exception as e:  # noqa: E722
            retries -= 1
            if retries < 0:
                raise SystemError from e
            time.sleep(1)

    stop_docker(container_name)


@pytest.fixture(scope="session")
def azure_fixture(azurite_credentials, docker_azurite):
    azure_storage = pytest.importorskip("azure.storage.blob")
    account_name, connection_string = azurite_credentials
    client = azure_storage.BlobServiceClient.from_connection_string(
        conn_str=connection_string
    )
    container_name = "data"
    client.create_container(container_name)

    yield f"az://{container_name}"



"""see upath/tests/conftest.py for fixtures
"""
import pytest  # noqa: F401

import fsspec
from upath import UPath
from upath.implementations.cloud import S3Path
from ..cases import BaseTests


class TestUPathS3(BaseTests):
    SUPPORTS_EMPTY_DIRS = False

    @pytest.fixture(autouse=True)
    def path(self, s3_fixture):
        path, anon, s3so = s3_fixture
        self.path = UPath(path, anon=anon, **s3so)
        self.anon = anon
        self.s3so = s3so

    def test_is_S3Path(self):
        assert isinstance(self.path, S3Path)

    def test_chmod(self):
        # todo
        pass

    def test_rmdir(self):
        dirname = "rmdir_test"
        mock_dir = self.path.joinpath(dirname)
        mock_dir.joinpath("test.txt").touch()
        mock_dir.rmdir()
        assert not mock_dir.exists()
        with pytest.raises(NotADirectoryError):
            self.path.joinpath("file1.txt").rmdir()

    def test_relative_to(self):
        assert "s3://test_bucket/file.txt" == str(
            UPath("s3://test_bucket/file.txt").relative_to(
                UPath("s3://test_bucket")
            )
        )

    def test_iterdir_root(self):
        client_kwargs = self.path._kwargs["client_kwargs"]
        bucket_path = UPath(
            "s3://other_test_bucket", client_kwargs=client_kwargs
        )
        bucket_path.mkdir(mode="private")

        (bucket_path / "test1.txt").touch()
        (bucket_path / "test2.txt").touch()

        for x in bucket_path.iterdir():
            assert x.name != ""
            assert x.exists()

    def test_touch_unlink(self):
        path = self.path.joinpath("test_touch.txt")
        path.touch()
        assert path.exists()
        path.unlink()
        assert not path.exists()

        # should raise FileNotFoundError since file is missing
        with pytest.raises(FileNotFoundError):
            path.unlink()

        # file doesn't exists, but missing_ok is True
        path.unlink(missing_ok=True)

    @pytest.mark.parametrize(
        "joiner", [["bucket", "path", "file"], "bucket/path/file"]
    )
    def test_no_bucket_joinpath(self, joiner):
        path = UPath("s3://", anon=self.anon, **self.s3so)
        path = path.joinpath(joiner)
        assert str(path) == "s3://bucket/path/file"

    def test_creating_s3path_with_bucket(self):
        path = UPath("s3://", bucket="bucket", anon=self.anon, **self.s3so)
        assert str(path) == "s3://bucket/"

    def test_iterdir_with_plus_in_name(self, s3_with_plus_chr_name):
        bucket, anon, s3so = s3_with_plus_chr_name
        p = UPath(
            f"s3://{bucket}/manual__2022-02-19T14:31:25.891270+00:00",
            anon=True,
            **s3so,
        )

        files = list(p.iterdir())
        assert len(files) == 1
        (file,) = files
        assert file == p.joinpath("file.txt")

    @pytest.mark.skip
    def test_makedirs_exist_ok_false(self):
        pass


@pytest.fixture
def s3_with_plus_chr_name(s3_server):
    anon, s3so = s3_server
    s3 = fsspec.filesystem("s3", anon=False, **s3so)
    bucket = "plus_chr_bucket"
    path = f"{bucket}/manual__2022-02-19T14:31:25.891270+00:00"
    s3.mkdir(path)
    s3.touch(f"{path}/file.txt")
    s3.invalidate_cache()
    try:
        yield bucket, anon, s3so
    finally:
        if s3.exists(bucket):
            for dir, _, keys in s3.walk(bucket):
                for key in keys:
                    s3.rm(f"{dir}/{key}")



import setuptools

from upath import __version__

with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name="universal_pathlib",
    version=__version__,
    author="Andrew Fulton",
    author_email="andrewfulton9@gmail.com",
    url="https://github.com/fsspec/universal_pathlib",
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    description="pathlib api extended to use fsspec backends",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
)



import pytest

from upath import UPath
from upath.implementations.memory import MemoryPath
from ..cases import BaseTests


class TestMemoryPath(BaseTests):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        if not local_testdir.startswith("/"):
            local_testdir = "/" + local_testdir
        path = f"memory:{local_testdir}"
        self.path = UPath(path)
        self.prepare_file_system()

    def test_is_MemoryPath(self):
        assert isinstance(self.path, MemoryPath)



import sys

import nox
from pathlib import Path


@nox.session()
def develop(session):
    session.install("flit")
    session.run(*"flit install -s".split())


@nox.session()
def black(session):
    session.install("black")
    session.run(*"black upath noxfile.py setup.py".split())


@nox.session()
def lint(session):
    session.install("flake8")
    session.run("flake8", "upath")


@nox.session()
def type_checking(session):
    session.install("mypy")
    session.run("mypy", "upath")


@nox.session()
def install(session):
    session.install(".")


@nox.session()
def smoke(session):
    if (3, 10) < sys.version_info <= (3, 11, 0, "final"):
        # workaround for missing aiohttp wheels for py3.11
        session.install(
            "aiohttp",
            "--no-binary",
            "aiohttp",
            env={"AIOHTTP_NO_EXTENSIONS": "1"},
        )

    session.install(
        "pytest",
        "adlfs",
        "aiohttp",
        "requests",
        "gcsfs",
        "s3fs",
        "moto[s3,server]",
        "webdav4[fsspec]",
        "wsgidav",
        "cheroot",
    )
    session.run(*"pytest --skiphdfs -vv upath".split())


@nox.session()
def build(session):
    session.install("flit")
    session.run(*"flit build".split())


@nox.session()
def rm_dirs(session):
    paths = ["build", "dist"]
    for path in paths:
        p = Path(path)
        if p.exists():
            session.run(*f"rm -rf {str(p)}".split())



import upath.core
import re


class _CloudAccessor(upath.core._FSSpecAccessor):
    def _format_path(self, path):
        """
        netloc has already been set to project via `CloudPath._from_parts`
        """
        return f"{path._url.netloc}/{path.path.lstrip('/')}"

    def mkdir(self, path, create_parents=True, **kwargs):
        if (
            not create_parents
            and not kwargs.get("exist_ok", False)
            and self._fs.exists(self._format_path(path))
        ):
            raise FileExistsError
        return super().mkdir(path, create_parents=create_parents, **kwargs)


# project is not part of the path, but is part of the credentials
class CloudPath(upath.core.UPath):
    _default_accessor = _CloudAccessor

    @classmethod
    def _from_parts(cls, args, url=None, **kwargs):
        if kwargs.get("bucket") and url is not None:
            bucket = kwargs.pop("bucket")
            url = url._replace(netloc=bucket)
        obj = super()._from_parts(args, url, **kwargs)
        return obj

    @classmethod
    def _from_parsed_parts(cls, drv, root, parts, url=None, **kwargs):
        if kwargs.get("bucket") and url is not None:
            bucket = kwargs.pop("bucket")
            url = url._replace(netloc=bucket)
        obj = super()._from_parsed_parts(drv, root, parts, url=url, **kwargs)
        return obj

    def _sub_path(self, name):
        """
        `gcsfs` and `s3fs` return the full path as `<bucket>/<path>` with
        `listdir` and `glob`. However, in `iterdir` and `glob` we only want the
        relative path to `self`.
        """
        sp = re.escape(self.path)
        subed = re.sub(f"^({self._url.netloc})?/?({sp}|{sp[1:]})/?", "", name)
        return subed

    def joinpath(self, *args):
        if self._url.netloc:
            return super().joinpath(*args)
        # handles a bucket in the path
        else:
            path = args[0]
            if isinstance(path, list):
                args_list = list(*args)
            else:
                args_list = path.split(self._flavour.sep)
            bucket = args_list.pop(0)
            self._kwargs["bucket"] = bucket
            return super().joinpath(*tuple(args_list))


class GCSPath(CloudPath):
    pass


class S3Path(CloudPath):
    pass


class AzurePath(CloudPath):
    pass
